import arcade

from arcade_curtains import Curtains, BaseScene


class CircleSprite(arcade.Sprite):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.textures = [
            arcade.make_soft_circle_texture(100, arcade.color.WHITE, 255, 255),
            arcade.make_soft_circle_texture(100, arcade.color.GREEN, 255, 255)
        ]
        self.texture = self.textures[0]
        self.center_x = 70
        self.center_y = 70

    def toggle_texture(self, *args, **kwargs):
        self.set_texture(int(not self.textures.index(self.texture)))


class SmallScene(BaseScene):
    def setup(self):
        self.sprites = arcade.SpriteList()
        self.actor = CircleSprite()

    def enter_scene(self, previous_scene):
        def callback():
            self.actor.animate(duration=.5, position=(70, 70), loop=True)

        self.actor.animate(position=(430, 430), duration=.5, callback=callback)
        self.sprites.append(self.actor)


class Window(arcade.Window):
    def __init__(self):
        super().__init__(500, 500, 'move')
        self.curtains = Curtains(self)
        self.curtains.add_scene('scene1', SmallScene())

    def setup(self):
        self.curtains.set_scene('scene1')


def run(window):
    window.setup()
    try:
        arcade.run()
    except KeyboardInterrupt:
        pass


run(Window())



import pytest
import arcade_curtains as c
from arcade_curtains.helpers import TriggerAttr, Widget
import operator as op
import arcade
from unittest import mock

orientations = {
    'topleft': (450, 725),
    'topright': (550, 725),
    'bottomleft': (450, 575),
    'bottomright': (550, 575)
}


@pytest.mark.parametrize('attr,expected', orientations.items())
def test_it_can_give_other_orientations(attr, expected):
    sprite = arcade.Sprite(center_x=500, center_y=650)
    sprite.width = 100
    sprite.height = 150
    assert getattr(sprite, attr) == expected


orientations = {
    'topleft': (450, 725),
    'topright': (550, 725),
    'bottomleft': (450, 575),
    'bottomright': (550, 575)
}


@pytest.mark.parametrize('attr,value', orientations.items())
def test_it_can_set_other_orientations(attr, value):
    sprite = arcade.Sprite()
    sprite.width = 100
    sprite.height = 150
    setattr(sprite, attr, value)
    assert sprite.position == (500, 650)


def test_it_can_observe_before_change():
    observable = c.helpers.ObservableSprite()
    handler = mock.Mock()
    observable.before_change('attr', handler)
    observable.attr = 10
    handler.assert_called_once()


def test_it_can_observe_after_change():
    observable = c.helpers.ObservableSprite()
    handler = mock.Mock()
    observable.after_change('attr', handler)
    observable.attr = 10
    handler.assert_called_once()


def test_it_can_observe_trigger_change():
    observable = c.helpers.ObservableSprite()
    handler = mock.Mock()
    health = TriggerAttr('health')
    observable.trigger(health > 10, handler)
    observable.health = 11
    handler.assert_called_once()


def test_it_can_anchor_sprites():
    anchor = c.helpers.AnchorPoint(50, 50)
    sprite1 = arcade.Sprite()
    sprite1.position = (100, 100)
    sprite2 = arcade.Sprite()
    sprite2.position = (200, 200)

    anchor.dock(sprite1)
    anchor.dock(sprite2)

    anchor.position = (300, 300)
    assert anchor.position == (300, 300)
    assert sprite1.position == (350, 350)
    assert sprite2.position == (450, 450)


def test_it_can_make_anchor_from_sprite():
    sprite1 = arcade.Sprite()
    sprite1.position = (100, 100)
    anchor = c.helpers.AnchorPoint.from_sprite(sprite1, orientation='position')
    assert anchor.position == (100, 100)


def test_it_can_delay_setting_an_attribute(sprite):
    sprite.health = 10
    delayed = c.helpers.delay_set_attribute(sprite, "health", 20)
    assert sprite.health == 10
    delayed()
    assert sprite.health == 20


@pytest.mark.parametrize('op', [op.lt, op.gt, op.eq, op.le, op.ge, op.ne])
def test_it_can_build_a_trigger(sprite, op):
    health = TriggerAttr('health')
    trigger = op(health, 10)
    with pytest.raises(NotImplementedError):
        trigger.check(10)
    sprite.health = 10
    trigger.bake(sprite)
    assert isinstance(trigger.check(10), bool)


def test_it_can_make_a_widget():
    class MyWidget(Widget):
        def setup_widget(self):
            sprite1 = arcade.Sprite()
            sprite1.position = (100, 100)
            sprite1.width = 100
            sprite1.height = 100

            sprite2 = arcade.Sprite()
            sprite2.position = (200, 100)
            sprite2.width = 100
            sprite2.height = 100

            sprite3 = arcade.Sprite()
            sprite3.position = (200, 200)
            sprite3.width = 100
            sprite3.height = 100

            sprite4 = arcade.Sprite()
            sprite4.position = (100, 200)
            sprite4.width = 100
            sprite4.height = 100

            self.sprites.extend([sprite1, sprite2, sprite3, sprite4])

    wdg = MyWidget()
    wdg.position = (300, 300)
    assert wdg.position == (300, 300)
    assert wdg.topleft == (200, 400)
    assert wdg.topright == (400, 400)
    assert wdg.bottomleft == (200, 200)
    assert wdg.bottomright == (400, 200)
    assert wdg.bottom == 200
    assert wdg.top == 400
    assert wdg.left == 200
    assert wdg.right == 400
    assert wdg.width == 200
    assert wdg.height == 200



import arcade
from example_helpers import CircleSprite, get_window, run

from arcade_curtains import BaseScene


class SmallScene(BaseScene):
    def setup(self):
        self.sprites = arcade.SpriteList()
        self.actor = CircleSprite()
        self.sprites.append(self.actor)

        handler_function = self.actor.toggle_texture
        self.events.hover(self.actor, handler_function)


window = get_window('hover_example', SmallScene)
run(window)



from scipy.interpolate import interp1d
import numpy as np
import arcade
import attr

TRACKED_ATTRIBUTES = [
    'center_x',
    'center_y',
    'position',
    'angle',
    'scale',
    'width',
    'height',
    'alpha',
    'left',
    'right',
    'bottom',
    'top',
]


def interp2d(speed, points):
    x_points, y_points = zip(*points)
    fx = interp1d(speed, x_points)
    fy = interp1d(speed, y_points)

    def func(val):
        return (float(fx(val)), float(fy(val)))

    return func


def _valid(value):
    if isinstance(value, (list, tuple, set)):
        return not all([np.isnan(a) for a in value])
    return not np.isnan(value)


class NoActiveSceneException(Exception):
    pass


class AnimationManager:
    def __init__(self):
        self.animations = []

    def fire(self, sprite, sequence):
        if isinstance(sequence, Sequence):
            animator = Animator(sprite, sequence)
        elif isinstance(sequence, Chain):
            animator = sequence
        else:
            raise ValueError('Cannot fire {}'.format(
                sequence.__class__.__name__))
        self.animations.append(animator)

    def _blip(self, delta):
        for animator in self.animations[:]:
            animator.blip(delta)
            if animator.finished:
                self.animations.remove(animator)

    def kill(self, sprite):
        for animator in self.animations[:]:
            if animator.kill(sprite):
                self.animations.remove(animator)


class AnimationManagerProxy:
    def __init__(self, sprite):
        self.sprite = sprite

    def __call__(self, sequence=None, manager=None, **kwargs):
        if isinstance(sequence, KeyFrame):
            sequence = self._sequence_from_keyframe(sequence, **kwargs)
        if not sequence and kwargs:
            sequence = self._sequence_from_kwargs(**kwargs)
        self._get_manager(manager).fire(self.sprite, sequence)

    def kill(self):
        curtains = arcade.get_window().curtains
        for scene in curtains.scenes.values():
            scene.animations.kill(self.sprite)

    def _sequence_from_kwargs(self, **kwargs):
        callback = kwargs.pop('callback', None)
        duration = kwargs.pop('duration', 1)
        loop = kwargs.pop('loop', False)
        sequence = Sequence(loop=loop)
        sequence.add_keyframe(
            keyframe=KeyFrame.from_sprite(self.sprite, kwargs.keys()))
        sequence.add_keyframe(duration, KeyFrame(**kwargs))

        if callback:
            sequence.add_callback(duration, callback)
        return sequence

    def _sequence_from_keyframe(self, keyframe, **kwargs):
        kwargs.update(keyframe.to_dict())
        return self._sequence_from_kwargs(**kwargs)

    def _get_manager(self, manager=None):
        if manager is None:
            curtains = arcade.get_window().curtains
            if not curtains.current_scene:
                msg = 'use curtains.set_scene before animating a sprite'
                raise NoActiveSceneException(msg)
            manager = curtains.current_scene.animations
        return manager


class SequenceIndex:
    def __init__(self, sequence, index):
        self.sequence = sequence
        self.index = index

    @property
    def frame(self):
        return self.sequence.keyframes.get(self.index)

    @frame.setter
    def frame(self, value):
        self.sequence.add_keyframe(self.index, value)

    @property
    def callback(self):
        return self.sequence.callbacks.get(self.index)

    @callback.setter
    def callback(self, value):
        self.sequence.add_callback(self.index, value)


@attr.s
class Sequence:
    default_interval = attr.ib(default=1)
    keyframes = attr.ib(attr.Factory(dict))
    callbacks = attr.ib(attr.Factory(dict))
    is_reversed = attr.ib(default=False)
    loop = attr.ib(default=False)

    def __iter__(self):
        return iter(self.keyframes.items())

    def __getitem__(self, key):
        return SequenceIndex(self, key)

    def __len__(self):
        return len(self.keyframes)

    def _sort(self, collection):
        for key in sorted(collection.keys()):
            collection[key] = collection.pop(key)

    def _clean_point_in_time(self, pit):
        if not pit:
            default = [-1 * self.default_interval]
            pit = (list(self.keyframes.keys())
                   or default)[-1] + self.default_interval
        return pit

    def add_keyframe(self,
                     point_in_time=None,
                     keyframe=None,
                     callback=None,
                     **kwargs):
        point_in_time = self._clean_point_in_time(point_in_time)
        if not keyframe and kwargs:
            keyframe = KeyFrame(**kwargs)
        self.keyframes[point_in_time] = keyframe
        if callback:
            self.callbacks[point_in_time] = callback
        self._sort(self.keyframes)
        self._sort(self.callbacks)

    def add_keyframes(self, *keyframes):
        for keyframe in keyframes:
            if isinstance(keyframe, dict):
                self.add_keyframe(**keyframe)
            else:
                self.add_keyframe(*keyframe)

    def add_callback(self, point_in_time, callback):
        self.callbacks[point_in_time] = callback
        self._sort(self.callbacks)

    @property
    def total_time(self):
        return list(self.keyframes.keys())[-1]

    def _to_point_in_times(self):
        duration = list(self.keyframes.keys())
        if len(duration) == 1:
            duration.append(duration[0] + 0.0001)
        all_points = list(zip(*[k.to_list() for k in self.keyframes.values()]))
        pits = {}
        interp_fn = {'position': interp2d}

        attributes = TRACKED_ATTRIBUTES
        for idx, attribute in enumerate(attributes):
            points = all_points[idx]
            if len(points) == 1:
                points = points * 2
            if self.is_reversed:
                points = list(reversed(points))

            include = [_valid(p) for p in points]
            if not any(include):
                continue

            fn = interp_fn.get(attribute, interp1d)

            pit = PointInTime(fn(duration, points), self.total_time)
            pits[attribute] = pit
        return pits

    @classmethod
    def from_sprite(cls, sprite):
        sequence = cls()
        sequence.add_keyframe(0, KeyFrame.from_sprite(sprite))
        return sequence


@attr.s
class KeyFrame:
    center_x = attr.ib(default=np.nan)
    center_y = attr.ib(default=np.nan)
    position = attr.ib(default=np.nan)
    angle = attr.ib(default=np.nan)
    scale = attr.ib(default=np.nan)
    width = attr.ib(default=np.nan)
    height = attr.ib(default=np.nan)
    alpha = attr.ib(default=np.nan)
    top = attr.ib(default=np.nan)
    bottom = attr.ib(default=np.nan)
    left = attr.ib(default=np.nan)
    right = attr.ib(default=np.nan)

    @classmethod
    def from_sprite(cls, sprite, only_keys=None):
        if not only_keys:
            only_keys = ['position', 'angle', 'width', 'height']
        kwargs = {}
        for key in only_keys:
            kwargs[key] = getattr(sprite, key)
        return cls(**kwargs)

    def to_list(self):
        return [getattr(self, attr) for attr in TRACKED_ATTRIBUTES]

    def to_dict(self):
        result = {}
        for key in TRACKED_ATTRIBUTES:
            value = getattr(self, key)
            if _valid(value):
                result[key] = value
        return result


class Animator:
    def __init__(self, sprite, sequence):
        self.sprite = sprite
        self.loop = sequence.loop
        self._elapsed_time = 0
        self._total_time = sequence.total_time
        self.pits = sequence._to_point_in_times()
        self.callbacks = iter(sequence.callbacks.items())
        self.upcoming_callback()

    def upcoming_callback(self):
        try:
            self._upcoming_callback = next(self.callbacks)
        except StopIteration:
            self._upcoming_callback = None

    def blip(self, delta):
        self._elapsed_time += delta
        for attrib, pit in self.pits.items():
            value = pit.point_at(self._elapsed_time)
            if _valid(value):
                setattr(self.sprite, attrib, value)

        self.check_for_callback()

        if self.finished:
            if self.loop:
                self._elapsed_time = 0

    def kill(self, sprite):
        if sprite is self.sprite:
            return True
        return False

    def check_for_callback(self):
        if not self._upcoming_callback:
            return
        time, callback = self._upcoming_callback
        if time <= self._elapsed_time:
            self.upcoming_callback()
            callback()

    @property
    def finished(self):
        return self._elapsed_time >= self._total_time

    @property
    def elapsed_time(self):
        return self._elapsed_time

    @property
    def remaining_time(self):
        return self._total_time - self._elapsed_time


class PointInTime:
    def __init__(self, fn, duration):
        self.fn = fn
        self.duration = duration

    def point_at(self, time):
        if time > self.duration:
            time = self.duration
        return self.fn(time)


class Chain:
    def __init__(self, loop=False, callback=None):
        self.loop = loop
        self.animations = []
        self.anim_queue = None
        self.current_animator = None
        self.finished = False
        self.callback = callback

    def add_sequences(self, *anims):
        for anim in anims:
            self.add_sequence(*anim)

    def add_sequence(self, *anim):
        self.animations.append(anim)

    def blip(self, delta):
        if not self.anim_queue:
            self.anim_queue = iter(self.animations)
            self._next_animator()

        if self.finished:
            return

        self.current_animator.blip(delta)
        if self.current_animator.finished:
            anim = self.current_animator
            mod = anim.elapsed_time - anim._total_time
            self._next_animator()
            self.current_animator.blip(mod)

    def kill(self, sprite):
        if self.anim_queue is None:
            return False

        if self.current_animator.sprite is sprite:
            return True

        anim_queue = []
        for anim_sprite, sequence in self.anim_queue:
            anim_queue.append((anim_sprite, sequence))
            if sprite is anim_sprite:
                return True
        self.anim_queue = iter(anim_queue)

        return False

    def _next_animator(self):
        try:
            sprite, sequence = next(self.anim_queue)
            animator = Animator(sprite, sequence)
            self.current_animator = animator
        except StopIteration:
            if self.callback:
                self.callback()
            if self.loop and self.animations:
                self.anim_queue = None
            else:
                self.finished = True



import sys
from unittest.mock import Mock, patch

import pytest

# Hard mocking pyglet.gl to account for the travis CI machines not having GLU
sys.modules['pyglet.gl'] = Mock()


@pytest.fixture(scope='function')
def sprite():
    points = ((0, 0), (0, 100), (100, 0), (100, 100))
    sprite = Mock(points=points)
    sprite.get_adjusted_hit_box.return_value = points
    return sprite



import arcade

from arcade_curtains import Curtains, BaseScene, AnchorPoint

import math


def rotate(origin, point, angle):
    ox, oy = origin
    px, py = point

    qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
    qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
    return qx, qy


def move_anchor(sprite, x, y, dx, dy, anchor):
    anchor.position = (x, y)


class CircleSprite(arcade.Sprite):
    def __init__(self, *args, **kwargs):
        size = kwargs.pop('size', 100)
        super().__init__(*args, **kwargs)
        self.textures = [
            arcade.make_soft_circle_texture(size, arcade.color.WHITE, 255,
                                            255),
        ]
        self.texture = self.textures[0]


class Planet(CircleSprite):
    def revolve(self, delta):
        self.position = rotate(self.center.position, self.position, delta)


class SmallScene(BaseScene):
    def setup(self):
        self.sprites = arcade.SpriteList()
        self.center_of_the_universe = CircleSprite(size=150)
        self.center_of_the_universe.position = (250, 250)
        anchor = AnchorPoint.from_sprite(self.center_of_the_universe,
                                         'position')

        planets = 6
        rad = math.radians(360 / planets)
        for i in range(planets):
            planet = Planet(size=50)
            planet.center = self.center_of_the_universe
            planet.position = rotate((250, 250), (150, 150), i * rad)
            anchor.dock(planet)
            self.sprites.append(planet)
            self.events.frame(planet.revolve)

        self.events.drag(self.center_of_the_universe, move_anchor,
                         {'anchor': anchor})
        self.sprites.append(self.center_of_the_universe)


class Window(arcade.Window):
    def __init__(self):
        super().__init__(500, 500, 'Anchor')
        self.curtains = Curtains(self)
        self.curtains.add_scene('scene1', SmallScene())

    def setup(self):
        self.curtains.set_scene('scene1')


def run(window):
    window.setup()
    try:
        arcade.run()
    except KeyboardInterrupt:
        pass


run(Window())



import math
import arcade

from arcade_curtains import Curtains, BaseScene, Sequence, KeyFrame
from example_helpers import TabGroup, RadioGroup, SquareSprite, Button

WHITE = arcade.color.WHITE
GREEN = arcade.color.GREEN
BRICK_RED = arcade.color.BRICK_RED
TEAL = arcade.color.TEAL


def to_valid_bound(value, minval, maxval):
    if value > maxval:
        value = maxval
    if value < minval:
        value = minval
    return value


class AnimationDetails:
    def __init__(self):
        self.keyframes = {}
        self.current_keyframe = None
        self.loop = False

    def set_current_keyframe(self, idx):
        self.current_keyframe = self.keyframes[idx]

    def add_keyframe(self, obj):
        idx = len(self.keyframes)
        self.keyframes[idx] = KeyFrame.from_sprite(obj)


class AppScene(BaseScene):
    def setup(self):
        self.details = AnimationDetails()
        self.obj = SquareSprite(size=200, color=WHITE)
        self.obj.position = (300, 460)

        self.sprites = arcade.SpriteList()

        self.setup_ui()
        self.setup_tabs()
        self.setup_radio()

        self.drag_proxy = DragProxy(self.details)
        self.drag_proxy.current_attribute = 'position'

        self.events.drag(self.obj, self.drag_proxy)
        self.sprites.append(self.obj)

    def setup_ui(self):
        self.frame = SquareSprite(size=600, color=TEAL)
        self.frame.position = (300, 460)
        self.sprites.append(self.frame)

        self.loop_checkbox = SquareSprite(color=BRICK_RED, size=40)
        self.loop_checkbox.position = (155, 80)
        self.sprites.append(self.loop_checkbox)

        def init_button(text, handler, x, y):
            btn = Button(text=text, events=self.events)
            self.events.click(btn, handler)
            btn.position = (x, y)
            self.sprites.append(btn)

        init_button('+1 Frame', self.add_keyframe, 70, 130)
        init_button('Loop', self.toggle_loop, 70, 80)
        init_button('Play', self.play, 70, 30)
        init_button('Stop', self.stop, 193, 30)

    def setup_tabs(self):
        tab_names = ['Position', 'Angle', 'Scale', 'Width', 'Height']
        self.tabs = TabGroup(self.sprites, self.events, tab_names, 600,
                             (300, 740), 4)
        self.tabs['Position'].active = True

        def set_mouse_to_value(sprite, *args, **kwargs):
            self.drag_proxy.current_attribute = sprite.text.lower()

        for name in tab_names:
            self.events.click(self.tabs[name], set_mouse_to_value)

    def setup_radio(self):
        self.keyframes_radio = RadioGroup(self.sprites, self.events, [], 560,
                                          (405, 130), 10)
        self.add_keyframe()

    def add_keyframe(self, *args, **kwargs):
        idx = len(self.details.keyframes)
        if idx >= 5:
            return
        self.keyframes_radio.add_button(str(idx))
        self.keyframes_radio[str(idx)].on_click()
        self.details.add_keyframe(self.obj)
        self.events.click(self.keyframes_radio[str(idx)], self.select_keyframe)
        self.details.set_current_keyframe(
            int(self.keyframes_radio.active.text))

    def select_keyframe(self, *args, **kwargs):
        self.details.set_current_keyframe(
            int(self.keyframes_radio.active.text))
        seq = Sequence()
        seq.add_keyframe(0, self.details.current_keyframe)
        self.obj.animate(seq)

    def toggle_loop(self, *args, **kwargs):
        self.details.loop = not self.details.loop
        self.loop_checkbox.set_texture(int(self.details.loop))

    def play(self, *args, **kwargs):
        seq = Sequence(loop=self.details.loop)
        seq.add_keyframes(*self.details.keyframes.items())
        self.obj.animate(seq)

    def stop(self, *args, **kwargs):
        self.curtains.current_scene.animations.animations = []


class DragProxy:
    def __init__(self, details):
        self.current_attribute = None
        self.details = details

    def __call__(self, *args, **kwargs):
        getattr(self, self.current_attribute)(*args, **kwargs)

    def angle(self, sprite, x, y, dx, dy):
        dx = x - sprite.center_x
        dy = y - sprite.center_y
        sprite.angle = (math.atan2(dy, dx) * 180 / math.pi)
        self.details.current_keyframe.angle = sprite.angle

    def position(self, sprite, x, y, dx, dy):
        new_pos = (x, y)
        sprite.position = new_pos
        self.details.current_keyframe.position = sprite.position

    def scale(self, sprite, x, y, dx, dy):
        val = (x - (x + dx)) * 0.05
        sprite.scale = to_valid_bound(sprite.scale + val, 1, 2)
        self.details.current_keyframe.scale = sprite.scale

    def width(self, sprite, x, y, dx, dy):
        val = (x - (x + dx)) * 0.5
        sprite.width = to_valid_bound(sprite.width + val, 100, 400)
        self.details.current_keyframe.width = sprite.width

    def height(self, sprite, x, y, dx, dy):
        val = (y - (y + dy)) * 0.5
        sprite.height = to_valid_bound(sprite.height + val, 100, 400)
        self.details.current_keyframe.height = sprite.height


class Window(arcade.Window):
    def __init__(self):
        super().__init__(600, 760, 'move')
        self.curtains = Curtains(self)
        self.curtains.add_scenes({
            'scene': AppScene(),
        })
        self.curtains.set_scene('scene')


Window()
arcade.run()



from unittest.mock import Mock

from arcade.application import MOUSE_BUTTON_LEFT
import pytest

from arcade_curtains import BaseScene, Curtains


class Scene(BaseScene):
    def setup(self):
        pass


@pytest.fixture(scope='function')
def curtains():
    window = Mock()
    curtains = Curtains(window)
    scene1 = Scene()
    scene2 = Scene()
    curtains.add_scenes({
        'scene1': scene1,
        'scene2': scene2,
    })
    curtains.set_scene('scene1')
    return curtains


def test_it_triggers_the_right_scene_event(sprite, curtains):
    scene1, scene2 = curtains.scenes.values()
    scene1.events.click(sprite, sprite.handler)
    scene2.events.click(sprite, sprite.handler)
    scene1.events.click(sprite, sprite.handler1)
    scene2.events.click(sprite, sprite.handler2)

    curtains.on_mouse_press(50, 50, MOUSE_BUTTON_LEFT, 0)
    curtains.on_mouse_release(50, 50, MOUSE_BUTTON_LEFT, 0)
    sprite.handler.assert_called_once()
    sprite.handler1.assert_called_once()

    curtains.set_scene('scene2')
    curtains.on_mouse_press(50, 50, MOUSE_BUTTON_LEFT, 0)
    curtains.on_mouse_release(50, 50, MOUSE_BUTTON_LEFT, 0)
    assert sprite.handler.call_count == 2
    sprite.handler2.assert_called_once()


def test_it_can_kill_a_sprite(sprite, curtains):
    def trigger():
        curtains.on_mouse_press(50, 50, MOUSE_BUTTON_LEFT, 0)
        curtains.on_mouse_release(50, 50, MOUSE_BUTTON_LEFT, 0)
        curtains.update(1)

    scene1 = curtains.scenes['scene1']
    scene1.events.down(sprite, sprite.handler)
    scene1.events.up(sprite, sprite.handler)
    scene1.events.frame(sprite.handler)
    trigger()

    assert sprite.handler.call_count == 3

    scene1.events.kill(sprite)

    trigger()
    assert sprite.handler.call_count == 3
    assert sprite not in scene1.events.all_sprites


def test_it_can_remove_a_sprite_handler(sprite, curtains):
    scene1 = curtains.scenes['scene1']
    scene1.events.down(sprite, sprite.handler)
    curtains.on_mouse_press(50, 50, MOUSE_BUTTON_LEFT, 0)

    assert sprite.handler.call_count == 1

    scene1.events.remove_down(sprite, sprite.handler)
    curtains.on_mouse_press(50, 50, MOUSE_BUTTON_LEFT, 0)

    assert sprite.handler.call_count == 1


def test_it_can_remove_a_handler(sprite, curtains):
    scene1 = curtains.scenes['scene1']
    scene1.events.frame(sprite.handler)
    curtains.update(1)

    assert sprite.handler.call_count == 1

    scene1.events.remove_frame(sprite.handler)
    curtains.update(1)

    assert sprite.handler.call_count == 1


def test_it_can_remove_a_handler_everywhere(sprite, curtains):
    scene1 = curtains.scenes['scene1']

    def trigger():
        scene1.events.trigger_before_draw()
        scene1.events.trigger_after_draw()
        curtains.update(1)

    scene1.events.before_draw(sprite.handler)
    scene1.events.after_draw(sprite.handler)
    scene1.events.frame(sprite.handler)
    trigger()

    assert sprite.handler.call_count == 3

    scene1.events.remove_from_all(sprite.handler)
    trigger()

    assert sprite.handler.call_count == 3



from collections import defaultdict
from functools import partial, wraps

import arcade

from .animation import AnimationManagerProxy


def delay_set_attribute(sprite, flag, value):
    """
    helper function for setting an attribute of an object as a callback in a
    sequence or chain.
    """

    def _delay_set_attribute():
        setattr(sprite, flag, value)

    return _delay_set_attribute


def modified_init(orig_init):
    @wraps(orig_init)
    def init(self, *args, **kwargs):
        orig_init(self, *args, **kwargs)
        self.animate = AnimationManagerProxy(self)

    return init


def modified_kill(orig_kill):
    @wraps(orig_kill)
    def kill(self, *args, **kwargs):
        orig_kill(self, *args, **kwargs)
        self.animate.kill()

    return kill


class PositionHelperMixin:
    @property
    def topleft(self):
        return (self.left, self.top)

    @property
    def topright(self):
        return (self.right, self.top)

    @property
    def bottomleft(self):
        return (self.left, self.bottom)

    @property
    def bottomright(self):
        return (self.right, self.bottom)

    @topleft.setter
    def topleft(self, position):
        self.left, self.top = position

    @topright.setter
    def topright(self, position):
        self.right, self.top = position

    @bottomleft.setter
    def bottomleft(self, position):
        self.left, self.bottom = position

    @bottomright.setter
    def bottomright(self, position):
        self.right, self.bottom = position


class CurtainsMeta(type):
    def __init__(self, name, bases, dct):
        super().__init__(name, bases, dct)
        self.__init__ = modified_init(self.__init__)
        self.kill = modified_kill(self.kill)


class Sprite(arcade.Sprite, PositionHelperMixin, metaclass=CurtainsMeta):
    pass


class TriggerAttr:
    def __init__(self, attribute):
        self.attribute = attribute

    def __eq__(self, other):
        return Trigger(self.attribute, '==', other)

    def __ne__(self, other):
        return Trigger(self.attribute, '!=', other)

    def __gt__(self, other):
        return Trigger(self.attribute, '>', other)

    def __ge__(self, other):
        return Trigger(self.attribute, '>=', other)

    def __lt__(self, other):
        return Trigger(self.attribute, '<', other)

    def __le__(self, other):
        return Trigger(self.attribute, '<=', other)


class Trigger:
    def __init__(self, attribute, op, value):
        """
        Nothing beats the speed of using relational operators, and getting
        attributes through dot accessing. Since this check might potentially
        be executed a good amount of times, we want to bake the check to be
        as fast as possible.
        """
        self.attribute = attribute
        self.op = op
        self.value = value

    def bake(self, obj):
        expression = "lambda s, x: s.{} {} {}".format(self.attribute, self.op,
                                                      self.value)
        self.check = partial(eval(expression), obj)

    def check(self, other):
        raise NotImplementedError(
            "Trigger is not yet baked. Call bake() first")


class ObservableSprite(Sprite):
    BEFORE_CHANGE = 0
    AFTER_CHANGE = 1
    TRIGGER = 2

    def __init__(self, *args, **kwargs):
        super().__setattr__(
            '_observe',
            defaultdict(lambda: [list(), list(), dict()]))
        super().__init__(*args, **kwargs)

    def __setattr__(self, key, value):
        if key in self._observe:
            old = getattr(self, key, None)
            handlers = self._observe[key]
            # Before handlers
            for handler in handlers[self.BEFORE_CHANGE]:
                handler(self, key, old, value)
            # Set the attribute
            super().__setattr__(key, value)
            # After handlers
            for handler in handlers[self.AFTER_CHANGE]:
                handler(self, key, old, value)
            # Triggers
            for check, handler in handlers[self.TRIGGER].items():
                if check(value):
                    handler()
        else:
            super().__setattr__(key, value)

    def observe(self, event_type, attribute, handler):
        self._observe[attribute][event_type].append(handler)

    def before_change(self, attribute, handler):
        self.observe(self.BEFORE_CHANGE, attribute, handler)

    def after_change(self, attribute, handler):
        self.observe(self.AFTER_CHANGE, attribute, handler)

    def trigger(self, trigger, handler):
        trigger.bake(self)
        self._observe[trigger.attribute][self.TRIGGER][trigger.check] = handler


class AnchorPoint:
    def __init__(self, center_x, center_y):
        self._center_x = center_x
        self._center_y = center_y
        self.boats = set()

    def dock(self, boat):
        self.boats.add(boat)

    @property
    def center_x(self):
        return self._center_x

    @property
    def center_y(self):
        return self._center_y

    @property
    def position(self):
        return self.center_x, self.center_y

    @center_x.setter
    def center_x(self, value):
        offset = value - self._center_x
        self._center_x = value
        for boat in self.boats:
            boat.center_x += offset

    @center_y.setter
    def center_y(self, value):
        offset = value - self._center_y
        self._center_y = value
        for boat in self.boats:
            boat.center_y += offset

    @position.setter
    def position(self, value):
        self.center_x, self.center_y = value

    @classmethod
    def from_sprite(cls, sprite, orientation):
        anchor = cls(*getattr(sprite, orientation))
        anchor.dock(sprite)
        return anchor


class Widget(PositionHelperMixin):
    def __init__(self, center_x=0, center_y=0, **kwargs):
        self.sprites = []
        self.setup_widget(**kwargs)
        self.anchor = AnchorPoint(
            center_x=self.center_x, center_y=self.center_y)
        for sprite in self.sprites:
            self.anchor.dock(sprite)
        self.anchor.position = (center_x, center_y)
        self.post_setup()

    def setup_widget(self):
        raise NotImplementedError()

    def post_setup(self):
        pass

    def _get_widget_bounds(self):
        return self._get_x_bounds, self._get_y_bounds

    def _get_y_bounds(self):
        if not self.sprites:
            return (0, 0)
        top = self.sprites[0].top
        bottom = self.sprites[0].bottom
        for sprite in self.sprites[1:]:
            top = max(top, sprite.top)
            bottom = min(bottom, sprite.bottom)
        return top, bottom

    def _get_x_bounds(self):
        if not self.sprites:
            return (0, 0)
        left = self.sprites[0].left
        right = self.sprites[0].right
        for sprite in self.sprites[1:]:
            left = min(left, sprite.left)
            right = max(right, sprite.right)
        return left, right

    @property
    def position(self):
        return self.center_x, self.center_y

    @position.setter
    def position(self, value):
        self.center_x, self.center_y = value

    @property
    def center_x(self):
        left, right = self._get_x_bounds()
        return ((right - left) / 2) + left

    @center_x.setter
    def center_x(self, value):
        self.anchor.center_x = value

    @property
    def center_y(self):
        top, bottom = self._get_y_bounds()
        return ((top - bottom) / 2) + bottom

    @center_y.setter
    def center_y(self, value):
        self.anchor.center_y = value

    @property
    def left(self):
        left, right = self._get_x_bounds()
        return left

    @left.setter
    def left(self, value):
        offset = value - self.left
        self.anchor.center_x += offset

    @property
    def right(self):
        left, right = self._get_x_bounds()
        return right

    @right.setter
    def right(self, value):
        offset = value - self.right
        self.anchor.center_x += offset

    @property
    def top(self):
        top, bottom = self._get_y_bounds()
        return top

    @top.setter
    def top(self, value):
        offset = value - self.top
        self.anchor.center_y += offset

    @property
    def bottom(self):
        top, bottom = self._get_y_bounds()
        return bottom

    @bottom.setter
    def bottom(self, value):
        offset = value - self.bottom
        self.anchor.center_y += offset

    @property
    def width(self):
        left, right = self._get_x_bounds()
        return right - left

    @property
    def height(self):
        top, bottom = self._get_y_bounds()
        return top - bottom

    def register(self, spritelist):
        for sprite in self.sprites:
            spritelist.append(sprite)


arcade.Sprite = Sprite
arcade.sprite.Sprite = Sprite



from unittest.mock import Mock

import pytest
from arcade.key import ESCAPE

from arcade_curtains.event import EventHandler


@pytest.mark.parametrize("event,triggers", [
    ('down', [('trigger_down', (50, 50))]),
    ('up', [('trigger_up', (50, 50))]),
    ('click', [('trigger_down', (50, 50)), ('trigger_up', (50, 50))]),
    ('hover', [('update', (50, 50)), ('update', (50, 50))]),
    ('out', [('update', (50, 50)), ('update', (-50, -50))]),
    ('drag', [('trigger_down', (50, 50)), ('trigger_drag', (50, 50, 1, 1))]),
])
def test_it_can_fire_sprite_events(sprite, event, triggers):
    ev = EventHandler()
    getattr(ev, event)(sprite, sprite.handler)
    for trigger, coords in triggers:
        getattr(ev, trigger)(*coords)
    sprite.handler.assert_called_once()


def test_it_can_fire_callbacks_with_args(sprite):
    ev = EventHandler()
    kwargs = {'default': True, 'attribute': 'health'}
    ev.down(sprite, sprite.handler, kwargs)
    ev.trigger_down(50, 50)
    sprite.handler.assert_called_with(sprite, 50, 50, **kwargs)


@pytest.mark.parametrize("event,triggers", [
    ('frame', [('trigger_frame', (50, ))]),
    ('before_draw', [('trigger_before_draw', ())]),
    ('after_draw', [('trigger_after_draw', ())]),
])
def test_it_can_fire_events(sprite, event, triggers):
    ev = EventHandler()
    getattr(ev, event)(sprite.handler)
    for trigger, args in triggers:
        getattr(ev, trigger)(*args)
    sprite.handler.assert_called_once()


def test_it_can_fire_keyboard_events(sprite):
    ev = EventHandler()
    ev.key(ESCAPE, sprite.handler)
    ev.trigger_key_press(ESCAPE)
    sprite.handler.assert_called_once()



import arcade

from arcade_curtains import Curtains, BaseScene, Sequence, KeyFrame, Chain


class CircleSprite(arcade.Sprite):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.textures = [
            arcade.make_soft_circle_texture(100, arcade.color.WHITE, 255, 255),
            arcade.make_soft_circle_texture(100, arcade.color.GREEN, 255, 255)
        ]
        self.texture = self.textures[0]
        self.center_x = 70
        self.center_y = 70

    def toggle_texture(self, *args, **kwargs):
        self.set_texture(int(not self.textures.index(self.texture)))


class SmallScene(BaseScene):
    def setup(self):
        self.sprites = arcade.SpriteList()
        self.actor1 = CircleSprite()
        self.actor2 = CircleSprite()
        self.actor2.position = (70, 430)
        self.sprites.append(self.actor1)
        self.sprites.append(self.actor2)

    def enter_scene(self, previous_scene):
        left1 = KeyFrame(position=self.actor1.position)
        left2 = KeyFrame(position=self.actor2.position)
        right1 = KeyFrame(position=(430, 70))
        right2 = KeyFrame(position=(430, 430))
        sequence1 = Sequence()
        sequence2 = Sequence()

        sequence1.add_keyframes((0, left1), (.5, right1), (1, left1))
        sequence2.add_keyframes((0, left2), (.5, right2), (1, left2))

        chain = Chain(loop=True)
        chain.add_sequences((self.actor1, sequence1), (self.actor2, sequence2))
        self.animations.fire(None, chain)


class Window(arcade.Window):
    def __init__(self):
        super().__init__(500, 500, 'move')
        self.curtains = Curtains(self)
        self.curtains.add_scene('scene1', SmallScene())

    def setup(self):
        self.curtains.set_scene('scene1')


def run(window):
    window.setup()
    try:
        arcade.run()
    except KeyboardInterrupt:
        pass


run(Window())



import arcade

from arcade_curtains import Curtains
from PIL import Image, ImageFont, ImageDraw

WHITE = arcade.color.WHITE
BLACK = arcade.color.BLACK
GREEN = arcade.color.GREEN
BRICK_RED = arcade.color.BRICK_RED
TEAL = arcade.color.TEAL
ROYAL_PURPLE = arcade.color.ROYAL_PURPLE
CORNFLOWER_BLUE = arcade.color.CORNFLOWER_BLUE


class BasicSprite(arcade.Sprite):
    texture_maker = None

    def __init__(self, *args, **kwargs):
        color = kwargs.pop('color', WHITE)
        self.size = kwargs.pop('size', 100)

        super().__init__(*args, **kwargs)
        self.texture = self.get_texture(self.size, color)
        self.textures = [
            self.texture,
            self.get_texture(self.size, TEAL),
            self.get_texture(self.size, CORNFLOWER_BLUE),
        ]

    def get_texture(self, size, color):
        raise NotImplementedError()


class SquareSprite(BasicSprite):
    def get_texture(self, size, color):
        return arcade.make_soft_square_texture(size, color, 255, 255)


class Button(SquareSprite):
    default_width = 115
    default_height = 40

    def __init__(self, text, events, *args, **kwargs):
        self.text = text
        self.button_height = kwargs.get('button_height', self.default_height)
        self.button_width = kwargs.get('button_width', self.default_width)

        events.hover(self, self.on_hover)
        events.out(self, self.on_out)
        events.down(self, self.down)
        events.up(self, self.up)

        super().__init__(*args, **kwargs)

    def get_texture(self, size, color):
        font = ImageFont.truetype("FreeMonoBold", 20)
        img = Image.new("RGBA", (self.button_width, self.button_height), color)
        draw = ImageDraw.Draw(img)
        draw.text((10, 10), self.text, font=font, align="left", fill=BLACK)
        name = '{}:{}:{}'.format('btn', color, self.text)
        return arcade.Texture(name, img)

    def on_hover(self, *args, **kwargs):
        self.set_texture(1)

    def on_out(self, *args, **kwargs):
        self.set_texture(0)

    def down(self, *args, **kwargs):
        self.set_texture(2)

    def up(self, *args, **kwargs):
        self.set_texture(1)


class ButtonGroup:
    button_class = None

    def __init__(self, spritelist, events, buttons, width, position, padding):
        self.buttons = {}
        self.spritelist = spritelist
        self.x, self.y = position
        self.padding = padding
        self.events = events
        self.tab_width = self.button_class.default_width
        self.start_x = (self.x - (width / 2)) + (self.tab_width / 2) + padding
        self.active = None

        for text in buttons:
            self.add_button(text)

    def add_button(self, text):
        btn = self.button_class(text=text, group=self, events=self.events)
        btn.center_x = self.start_x + (
            (self.tab_width + self.padding) * len(self.buttons))
        btn.center_y = self.y
        self.buttons[text] = btn
        self.spritelist.append(btn)

    def __iter__(self):
        return iter(self.buttons.values())

    def __getitem__(self, key):
        return self.buttons[key]


class TabButton(Button):
    def __init__(self, group, *args, **kwargs):
        self._active = False
        self.group = group
        super().__init__(*args, **kwargs)

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, value):
        self._active = value
        self.on_out()

    def on_out(self, *args, **kwargs):
        if self.active:
            self.set_texture(1)
        else:
            self.set_texture(0)

    def on_click(self, *args, **kwargs):
        self.group.active = self
        for btn in self.group:
            btn.active = btn == self


class RadioButton(Button):
    default_width = 40
    border_width = 7

    def __init__(self, group, *args, **kwargs):
        self.group = group
        super().__init__(*args, **kwargs)

    def on_click(self, *args, **kwargs):
        try:
            self.group.selected_border.animate(
                position=self.position,
                width=self.width + self.border_width,
                height=self.height + self.border_width,
                duration=.1)
        except Exception:
            self.group.selected_border.position = self.position
            self.group.selected_border.width = self.width + self.border_width
            self.group.selected_border.height = self.height + self.border_width
        self.group.active = self


class RadioGroup(ButtonGroup):
    button_class = RadioButton

    def __init__(self, spritelist, *args, **kwargs):
        self.selected_border = SquareSprite(color=BRICK_RED)
        spritelist.append(self.selected_border)
        super().__init__(spritelist, *args, **kwargs)


class TabGroup(ButtonGroup):
    button_class = TabButton


class CircleSprite(arcade.Sprite):
    def __init__(self, color=arcade.color.WHITE, size=100, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.textures = [
            arcade.make_soft_circle_texture(size, color, 255, 255),
            arcade.make_soft_circle_texture(size, arcade.color.GREEN, 255, 255)
        ]
        self.texture = self.textures[0]
        self.center_x = 70
        self.center_y = 70

    def toggle_texture(self, *args, **kwargs):
        self.set_texture(int(not self.textures.index(self.texture)))


def get_window(name, Scene):
    class Window(arcade.Window):
        def __init__(self):
            super().__init__(140, 140, name)
            self.curtains = Curtains(self)
            self.curtains.add_scene('scene1', Scene())

        def setup(self):
            self.curtains.set_scene('scene1')

    return Window()


def run(window):
    window.setup()
    try:
        arcade.run()
    except KeyboardInterrupt:
        pass



import arcade
from scipy.interpolate import interp1d

from arcade_curtains import Curtains
from arcade_curtains.scene import BaseScene

SCREEN_WIDTH = 620
SCREEN_HEIGHT = 260
SCREEN_TITLE = "Play"
PADDING = 20

WHITE = arcade.color.WHITE
BLACK = arcade.color.BLACK
GREEN = arcade.color.GREEN
RED = arcade.color.RED


class BasicSprite(arcade.Sprite):
    texture_maker = None

    def __init__(self, *args, **kwargs):
        color = kwargs.pop('color', BLACK)
        self.size = kwargs.pop('size', 100)

        super().__init__(*args, **kwargs)
        self.texture = self.get_texture(self.size, color)
        self.textures = [
            self.texture,
            self.get_texture(self.size, GREEN),
            self.get_texture(self.size, RED),
        ]

    def get_texture(self, size, color):
        raise NotImplementedError()


class SquareSprite(BasicSprite):
    def get_texture(self, size, color):
        return arcade.make_soft_square_texture(size, color, 255, 255)


class CircleSprite(BasicSprite):
    def get_texture(self, size, color):
        return arcade.make_soft_circle_texture(size, color, 255, 255)

    def green(self, *args):
        self.set_texture(1)

    def black(self, *args):
        self.set_texture(0)


class MoveAnimator:
    def __init__(self, sprite, x, y, events):
        self.speed = .1
        self.sprite = sprite
        self.elapsed_time = 0
        self.events = events

        x_points = [sprite.center_x, x]
        y_points = [sprite.center_y, y]
        self.x_path = interp1d([0, self.speed], x_points)
        self.y_path = interp1d([0, self.speed], y_points)

        self.events.frame(self.move)

    def move(self, delta):
        self.elapsed_time += delta
        if self.elapsed_time > self.speed:
            self.elapsed_time = self.speed
            # The animation has ended, remove it from the event handler
            self.events.remove_frame(self.move)

        self.sprite.center_x = float(self.x_path(self.elapsed_time))
        self.sprite.center_y = float(self.y_path(self.elapsed_time))


class CharacterDevelopment(BaseScene):
    next_scene = 'scene2'
    primary_color = WHITE
    secondary_color = BLACK

    def setup(self):
        # set up button drawer
        self.labels = []
        self.buttons = arcade.SpriteList()
        self.actors = arcade.SpriteList()

        # set up buttons
        btn_add_actor = SquareSprite(
            center_x=60, center_y=60, color=self.secondary_color)
        btn_next_scene = SquareSprite(
            center_x=180, center_y=60, color=self.secondary_color)
        self.buttons.append(btn_add_actor)
        self.buttons.append(btn_next_scene)

        # set up button events
        self.events.click(btn_next_scene,
                          lambda *x: self.curtains.set_scene(self.next_scene))
        self.events.click(btn_add_actor, self.add_actor)

        self.add_label(btn_add_actor, "Add\nActor", self.primary_color)
        self.add_label(btn_next_scene, "Next\nScene", self.primary_color)

    def add_label(self, attach_to, text, color):
        x = attach_to.center_x - (attach_to.width / 2) + 10
        y = attach_to.center_y

        def draw_label(*args, **kwargs):
            arcade.draw_text(text, x, y, color, 14)

        self.events.after_draw(draw_label)

    def enter_scene(self, previous_scene):
        arcade.set_background_color(self.primary_color)

    def add_actor(self, sprite, x, y):
        actor = CircleSprite(color=self.secondary_color)
        x = SCREEN_WIDTH / 2
        y = SCREEN_HEIGHT - (actor.height / 2) - PADDING
        if len(self.actors):
            neighbour = self.actors.sprite_list[-1]
            x = neighbour.center_x + actor.width + PADDING
        actor.center_x = x
        actor.center_y = y
        self.actors.append(actor)

        self.events.hover(actor, actor.green)
        self.events.out(actor, actor.black)

        if len(self.actors) == 1:
            return

        if len(self.actors) == 5:
            # We don't want any more actors, the stage is getting too crowded
            self.events.remove_click(sprite, self.add_actor)
            sprite.set_texture(2)

        # Ass some nice animations
        offset = (actor.width + PADDING) / 2
        for actor in self.actors:
            MoveAnimator(actor, actor.center_x - offset, actor.center_y,
                         self.events)


class PlotTwist(CharacterDevelopment):
    next_scene = 'scene3'
    previous_scene = 'scene1'
    primary_color = BLACK
    secondary_color = WHITE

    def setup(self):
        super().setup()
        btn_previous_scene = SquareSprite(
            center_x=300, center_y=60, color=self.secondary_color)
        self.events.click(
            btn_previous_scene,
            lambda *x: self.curtains.set_scene(self.previous_scene))
        self.buttons.append(btn_previous_scene)

        self.add_label(btn_previous_scene, "Previous\nScene",
                       self.primary_color)


class DeusExMachina(BaseScene):
    def setup(self):
        self.buttons = arcade.SpriteList()
        btn_scene1 = SquareSprite(center_x=60, center_y=60)
        btn_scene2 = SquareSprite(center_x=180, center_y=60)
        self.events.click(btn_scene1,
                          lambda *x: self.curtains.set_scene('scene1'))
        self.events.click(btn_scene2,
                          lambda *x: self.curtains.set_scene('scene2'))
        self.buttons.append(btn_scene1)
        self.buttons.append(btn_scene2)

        self.add_label(btn_scene1, "Go To\nScene1", WHITE)
        self.add_label(btn_scene2, "Go To\nScene2", WHITE)

    def enter_scene(self, previous_scene):
        arcade.set_background_color(GREEN)

    add_label = CharacterDevelopment.add_label


class Window(arcade.Window):
    def __init__(self):
        super().__init__(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_TITLE)
        self.curtains = Curtains(self)
        self.curtains.add_scenes({
            'scene1': CharacterDevelopment(),
            'scene2': PlotTwist(),
            'scene3': DeusExMachina()
        })

    def setup(self):
        self.curtains.set_scene('scene1')


if __name__ == "__main__":
    window = Window()
    window.setup()
    try:
        arcade.run()
    except KeyboardInterrupt:
        pass



from functools import wraps

import arcade
from pyglet import clock

from .scene import BaseScene  # noqa
from .animation import Sequence, KeyFrame, Chain  # noqa
from .helpers import (  # noqa
    PositionHelperMixin, Sprite, ObservableSprite, AnchorPoint, Widget)


def bind(fn, window_fn):
    @wraps(window_fn)
    def decorator(*args, **kwargs):
        res = window_fn(*args, **kwargs)
        fn(*args, **kwargs)
        return res

    return decorator


class Curtains:
    def __init__(self, window=None):
        self.current_scene = None
        self.window = None
        self.scenes = {}
        if window:
            self.bind(window)

    def add_scene(self, name, scene):
        self.scenes[name] = scene
        scene._bind(self.window, self)

    def add_scenes(self, scenes):
        for name, scene in scenes.items():
            self.add_scene(name, scene)

    def set_scene(self, name):
        if self.current_scene:
            self.current_scene.leave_scene(self.scenes[name])
        previous_scene = self.current_scene
        self.current_scene = self.scenes[name]
        self.current_scene.enter_scene(previous_scene)

    def bind(self, window=None):
        if window:
            self.window = window
            self.window.curtains = self
            for scene in self.scenes.values():
                scene._bind(self.window, self)
        if not self.window:
            msg = "Curtains instance is not yet bound to an Arcade window"
            raise Exception(msg)

        clock.unschedule(window.update)
        members = {
            'update': self.update,
            'on_draw': self.on_draw,
            'on_mouse_motion': self.on_mouse_motion,
            'on_mouse_press': self.on_mouse_press,
            'on_mouse_release': self.on_mouse_release,
            'on_mouse_drag': self.on_mouse_drag,
            'on_key_press': self.on_key_press
        }
        for member, fn in members.items():
            window_fn = getattr(window, member)
            setattr(window, member, bind(fn, window_fn))
        # clock.schedule(window.update)
        clock.schedule_interval(window.update, 1 / 60)

    def on_draw(self):
        arcade.start_render()

        self.current_scene.events.trigger_before_draw()
        self.current_scene.draw()
        self.current_scene.events.trigger_after_draw()

    def update(self, delta_time):
        self.current_scene.events.trigger_frame(delta_time)

    def on_mouse_motion(self, x, y, dx, dy):
        self.current_scene.events.update(x, y)

    def on_mouse_press(self, x, y, button, modifiers):
        if button == arcade.MOUSE_BUTTON_LEFT:
            self.current_scene.events.trigger_down(x, y)

    def on_mouse_release(self, x, y, button, modifiers):
        if button == arcade.MOUSE_BUTTON_LEFT:
            self.current_scene.events.trigger_up(x, y)

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        self.current_scene.events.trigger_drag(x, y, dx, dy)

    def on_key_press(self, key, modifiers):
        if not modifiers:
            self.current_scene.events.trigger_key_press(key)



from collections import defaultdict
from enum import Enum
from functools import partialmethod
from unittest.mock import Mock

import arcade


class SpriteEvent(Enum):
    CLICK = 1
    OUT = 2
    HOVER = 3
    DOWN = 4
    UP = 5
    DRAG = 6


class Event(Enum):
    FRAME = 7
    BEFORE_DRAW = 8
    AFTER_DRAW = 9


EMPTY_SPRITE = Mock()


def run_handlers(handlers, *args, **kwargs):
    for handler, handler_kwargs in handlers:
        kwargs.update(handler_kwargs)
        handler(*args, **kwargs)


class EventHandler(object):
    def __init__(self):
        self.sprite_handlers = defaultdict(lambda: defaultdict(list))
        self.handlers = defaultdict(list)

        self.all_sprites = []

        self._previous_hover = EMPTY_SPRITE
        self._previous_down = EMPTY_SPRITE

        self.current_x = 0
        self.current_y = 0

    def update(self, x, y):
        self.current_x = x
        self.current_y = y
        current_hover = self.get_sprite_at(x, y)
        if current_hover is not self._previous_hover:
            self.trigger_hover_out(self._previous_hover, x, y)
            self.trigger_hover(current_hover, x, y)
            self._previous_hover = current_hover

    def trigger_hover_out(self, sprite, x, y):
        handlers = self.sprite_handlers[SpriteEvent.OUT].get(sprite, [])
        run_handlers(handlers, sprite, x, y)

    def trigger_hover(self, sprite, x, y):
        handlers = self.sprite_handlers[SpriteEvent.HOVER].get(sprite, [])
        run_handlers(handlers, sprite, x, y)

    def trigger_down(self, x, y):
        current_down = self.get_sprite_at(x, y)
        handlers = self.sprite_handlers[SpriteEvent.DOWN].get(current_down, [])
        run_handlers(handlers, current_down, x, y)
        self._previous_down = current_down

    def trigger_up(self, x, y):
        current_up = self.get_sprite_at(x, y)
        handlers = self.sprite_handlers[SpriteEvent.UP].get(current_up, [])
        run_handlers(handlers, current_up, x, y)
        if current_up is self._previous_down:
            self.trigger_click(current_up, x, y)
        self._previous_down = EMPTY_SPRITE

    def trigger_click(self, sprite, x, y):
        handlers = self.sprite_handlers[SpriteEvent.CLICK].get(sprite, [])
        run_handlers(handlers, sprite, x, y)

    def trigger_drag(self, x, y, dx, dy):
        drag_sprites = self.sprite_handlers[SpriteEvent.DRAG]
        handlers = drag_sprites.get(self._previous_down, [])
        run_handlers(handlers, self._previous_down, x, y, dx, dy)

    def trigger_frame(self, delta_time):
        run_handlers(self.handlers[Event.FRAME], delta_time)

    def trigger_before_draw(self):
        run_handlers(self.handlers[Event.BEFORE_DRAW])

    def trigger_after_draw(self):
        run_handlers(self.handlers[Event.AFTER_DRAW])

    def trigger_key_press(self, key):
        run_handlers(self.handlers.get(key, []))

    def get_sprite_at(self, *coords):
        sprites = arcade.SpriteList()
        sprites.sprite_list = self.all_sprites
        sprites = arcade.get_sprites_at_point(coords, sprites)
        if sprites:
            return max(sprites)
        return EMPTY_SPRITE

    def add_sprite_event(self, event_type, sprite, handler_function,
                         kwargs={}):
        self.all_sprites.append(sprite)
        self.all_sprites = list(set(self.all_sprites))
        self.sprite_handlers[event_type][sprite].append((handler_function,
                                                         kwargs))

    def add_event(self, event_type, handler_function, kwargs={}):
        if handler_function not in self.handlers[event_type]:
            self.handlers[event_type].append((handler_function, kwargs))

    def remove_sprite_event(self, event_type, sprite, handler):
        if self.sprite_handlers[event_type].get(sprite, None):
            for details in self.sprite_handlers[event_type][sprite]:
                if details[0] == handler:
                    self.sprite_handlers[event_type][sprite].remove(details)

    def remove_event(self, event_type, handler):
        for details in self.handlers[event_type]:
            if details[0] == handler:
                self.handlers[event_type].remove(details)

    def remove_from_all(self, handler):
        for event_type, handlers in self.handlers.items():
            self.remove_event(event_type, handler)

    def kill(self, sprite):
        for event in SpriteEvent:
            self.sprite_handlers[event].pop(sprite, None)
        for attribute_name in dir(sprite):
            attr = getattr(sprite, attribute_name)
            if callable(attr):
                for event in Event:
                    self.remove_event(event, attr)
        if sprite in self.all_sprites:
            self.all_sprites.remove(sprite)

    click = partialmethod(add_sprite_event, SpriteEvent.CLICK)
    hover = partialmethod(add_sprite_event, SpriteEvent.HOVER)
    out = partialmethod(add_sprite_event, SpriteEvent.OUT)
    down = partialmethod(add_sprite_event, SpriteEvent.DOWN)
    up = partialmethod(add_sprite_event, SpriteEvent.UP)
    drag = partialmethod(add_sprite_event, SpriteEvent.DRAG)
    frame = partialmethod(add_event, Event.FRAME)
    before_draw = partialmethod(add_event, Event.BEFORE_DRAW)
    after_draw = partialmethod(add_event, Event.AFTER_DRAW)

    remove_click = partialmethod(remove_sprite_event, SpriteEvent.CLICK)
    remove_hover = partialmethod(remove_sprite_event, SpriteEvent.HOVER)
    remove_out = partialmethod(remove_sprite_event, SpriteEvent.OUT)
    remove_down = partialmethod(remove_sprite_event, SpriteEvent.DOWN)
    remove_up = partialmethod(remove_sprite_event, SpriteEvent.UP)
    remove_drag = partialmethod(remove_sprite_event, SpriteEvent.DRAG)
    remove_frame = partialmethod(remove_event, Event.FRAME)
    remove_before_draw = partialmethod(remove_event, Event.BEFORE_DRAW)
    remove_after_draw = partialmethod(remove_event, Event.AFTER_DRAW)

    key = add_event
    remove_key = remove_event



from arcade_curtains import animation as a
import arcade
import numpy as np
from unittest import mock


def sprite_instance():
    s = arcade.Sprite()
    s.center_x = 10
    s.center_y = 20
    s.position = (10, 20)
    s.angle = 50
    s.width = 100
    s.height = 200
    s.scale = 1
    return s


def test_it_can_interp2d():
    points = ([0, 0], [5, 10], [20, 40])
    f = a.interp2d([0, 1, 2], points)
    assert f(.5) == (2.5, 5.)
    assert f(1.5) == (12.5, 25.)


def test_it_can_patch_arcade_sprite():
    assert isinstance(arcade.Sprite().animate, a.AnimationManagerProxy)


def test_it_can_patch_subclassed_arcade_sprite():
    class SomeSubClass(arcade.Sprite):
        pass

    assert isinstance(SomeSubClass().animate, a.AnimationManagerProxy)


def test_it_can_give_a_correcT_point_in_time():
    pit = a.PointInTime(a.interp1d([0, 1], [0, 10]), 1)
    assert pit.point_at(.1) == 1
    assert pit.point_at(1) == 10
    assert pit.point_at(10) == 10


def test_it_can_make_a_keyframe_from_sprite():
    s = sprite_instance()
    k = a.KeyFrame.from_sprite(s)
    assert k.to_dict() == {
        'position': (10, 20),
        'angle': 50,
        'width': 100,
        'height': 200,
    }


def test_it_can_make_a_keyframe_from_sprite_with_target_keys():
    s = sprite_instance()
    k = a.KeyFrame.from_sprite(s, only_keys=['center_x', 'center_y', 'scale'])
    assert k.to_dict() == {
        'center_x': 10,
        'center_y': 20,
        'scale': 1,
    }


def test_it_can_transform_a_keyframe_to_list():
    s = sprite_instance()
    k = a.KeyFrame.from_sprite(s)
    expected = [
        np.nan,
        np.nan,
        (10, 20),
        50,
        np.nan,
        100,
        200,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
        np.nan,
    ]
    assert k.to_list() == expected


def test_it_can_add_a_keyframe_to_a_sequence():
    s = a.Sequence()
    k = a.KeyFrame()
    s.add_keyframe(0, k)
    assert id(s[0].frame) == id(k)


def test_it_can_add_a_keyframes_to_a_sequence():
    s = a.Sequence()
    k1 = a.KeyFrame()
    k2 = a.KeyFrame()
    some_callback = lambda: 1
    s.add_keyframes((0, k1), (1, k2, some_callback))
    assert id(s[0].frame) == id(k1)
    assert id(s[1].frame) == id(k2)
    assert s[1].callback == some_callback


def test_it_can_make_a_sequence_from_a_sprite():
    s = a.Sequence().from_sprite(sprite_instance())
    assert len(s) == 1
    assert s[0].frame.position == (10, 20)


def test_it_can_generate_points_in_time_from_a_sequence():
    s = a.Sequence()
    k1 = a.KeyFrame(position=(10, 10), angle=0, height=100)
    k2 = a.KeyFrame(position=(10, 100), angle=100, height=200)
    s.add_keyframes((0, k1), (1, k2))
    pits = s._to_point_in_times()
    assert len(pits) == 3
    assert list(pits.keys()) == ['position', 'angle', 'height']
    assert pits['position'].point_at(.5) == (10, 55)
    assert pits['angle'].point_at(.5) == 50
    assert pits['height'].point_at(.5) == 150
    assert pits['height'].point_at(1) == 200
    assert pits['height'].point_at(10) == 200


def test_it_can_animate_a_sprite_over_time():
    s = a.Sequence()
    k1 = a.KeyFrame(position=(10, 10), angle=0, height=100)
    k2 = a.KeyFrame(position=(10, 100), angle=100, height=200)
    s.add_keyframes((0, k1), (1, k2))
    inst = sprite_instance()

    anim = a.Animator(inst, s)
    anim.blip(.5)
    assert inst.position == (10, 55)
    assert not anim.finished
    anim.blip(1)
    assert inst.position == (10, 100)
    assert anim.finished


def test_it_can_execute_a_callback_at_a_given_time():
    s = a.Sequence()
    k1 = a.KeyFrame(position=(10, 10))
    k2 = a.KeyFrame(position=(10, 100))
    s.add_keyframes((0, k1), (1, k2))
    callback = mock.Mock().some_callback
    s.add_callback(.5, callback)
    inst = sprite_instance()
    anim = a.Animator(inst, s)
    anim.blip(.4)
    callback.assert_not_called()
    anim.blip(.2)
    callback.assert_called_once()
    anim.blip(.4)
    callback.assert_called_once()
    assert not anim._upcoming_callback


def test_it_can_make_a_sequence_from_kwargs():
    proxy = a.AnimationManagerProxy(arcade.Sprite())
    seq = proxy._sequence_from_kwargs(
        position=(100, 400), duration=10, callback=lambda: None, loop=True)
    assert seq.loop
    assert seq.total_time == 10
    assert seq.callbacks[10]
    assert len(seq) == 2
    assert seq[10].frame.position == (100, 400)


def test_it_can_make_a_sequence_from_keyframe():
    proxy = a.AnimationManagerProxy(arcade.Sprite())
    seq = proxy._sequence_from_keyframe(
        a.KeyFrame(position=(100, 400)),
        duration=10,
        callback=lambda: None,
        loop=True)
    assert seq.loop
    assert seq.total_time == 10
    assert seq.callbacks[10]
    assert len(seq) == 2
    assert seq[10].frame.position == (100, 400)


@mock.patch('arcade_curtains.animation.AnimationManagerProxy._get_manager')
def test_it_can_fire_an_animation(mngr):
    manager = mock.Mock()
    mngr.return_value = manager
    proxy = a.AnimationManagerProxy(arcade.Sprite())
    proxy(position=(100, 400), )
    manager.fire.assert_called_once()


def test_it_can_play_animations():
    s = a.Sequence()
    k1 = a.KeyFrame(position=(10, 10))
    k2 = a.KeyFrame(position=(10, 100))
    s.add_keyframes((0, k1), (1, k2))
    inst = arcade.Sprite()
    manager = a.AnimationManager()
    manager.fire(inst, s)
    assert len(manager.animations) == 1
    manager._blip(.5)
    assert len(manager.animations) == 1
    assert inst.position == (10, 55)
    manager._blip(2)
    assert len(manager.animations) == 0
    assert inst.position == (10, 100)


def test_it_can_chain_animations():
    def make_sequence():
        s = a.Sequence()
        k1 = a.KeyFrame(position=(10, 10))
        k2 = a.KeyFrame(position=(10, 100))
        s.add_keyframes((0, k1), (1, k2))
        return s

    seq1 = make_sequence()
    seq2 = make_sequence()
    inst1 = arcade.Sprite()
    inst2 = arcade.Sprite()
    chain = a.Chain()
    chain.add_sequence(inst1, seq1)
    chain.add_sequence(inst2, seq2)
    chain.blip(0)
    assert inst1.position == (10, 10)
    assert inst2.position == (0, 0)
    chain.blip(.5)
    assert inst1.position == (10, 55)
    assert inst2.position == (0, 0)
    chain.blip(1)
    assert inst1.position == (10, 100)
    assert inst2.position == (10, 55)
    chain.blip(1)
    assert inst1.position == (10, 100)
    assert inst2.position == (10, 100)
    assert chain.finished


def test_it_can_call_animation_kill_on_sprite_kill():
    sprite = arcade.Sprite()
    animate = mock.Mock()
    sprite.animate = animate
    sprite.kill()
    animate.kill.assert_called_once()


@mock.patch('arcade.get_window')
def test_it_can_kill_a_sprite_in_animation_manager(window):
    sprite = mock.Mock(position=(0, 0))
    manager = a.AnimationManager()
    proxy = a.AnimationManagerProxy(sprite)
    scene = mock.Mock(animations=manager)
    window().curtains = mock.Mock(scenes={'name': scene})

    proxy(manager=manager, position=(100, 400), duration=10)
    assert manager.animations[0].sprite is sprite
    proxy.kill()
    assert not manager.animations


def test_it_doesnt_break_when_killing_on_unstarted_chain():
    chain = a.Chain()
    assert not chain.kill(mock.Mock())


def test_it_can_kill_from_chain_if_current_animator_has_sprite():
    chain = a.Chain()
    chain.anim_queue = 1
    sprite = mock.Mock()
    current_animator = mock.Mock(sprite=sprite)
    chain.current_animator = current_animator
    assert chain.kill(sprite)


def test_it_can_kill_from_chain_if_sprite_is_in_future_animations():
    chain = a.Chain()
    sprite = mock.Mock()
    chain.anim_queue = iter([(sprite, None)])
    current_animator = mock.Mock(sprite=mock.Mock())
    chain.current_animator = current_animator
    assert chain.kill(sprite)


def test_it_can_reconstruct_anim_queue_on_invalid_kill():
    chain = a.Chain()
    anim = (mock.Mock(), None)
    chain.anim_queue = iter([anim])
    current_animator = mock.Mock(sprite=mock.Mock())
    chain.current_animator = current_animator
    assert not chain.kill(mock.Mock())
    assert list(chain.anim_queue) == [anim]



import arcade
import random

from arcade_curtains import (Curtains, BaseScene, ObservableSprite, KeyFrame,
                             Sequence, Widget)
from arcade_curtains.helpers import delay_set_attribute, TriggerAttr
from example_helpers import SquareSprite, Button, CircleSprite

STATE = None


# State management
def get_state():
    global STATE
    if STATE is None:
        STATE = State()
    return STATE


class State:
    IDLE = 1
    WAITING = 2

    def __init__(self):
        self.current_state = self.IDLE
        self.handlers = []

    def __eq__(self, other):
        return self.current_state == other

    def set_waiting(self):
        self.current_state = self.WAITING

    def set_idle(self):
        self.current_state = self.IDLE

    def update(self, delta):
        if self == State.WAITING:
            return
        for handler in self.handlers:
            handler(delta)

    def add_handler(self, handler):
        self.handlers.append(handler)


# function to allow readability
def play_animation(args):
    sprite, seq = args
    sprite.animate(seq)


# animation maker functions
def powerup(sprite):
    flash = SquareSprite()
    flash.color = arcade.color.GOLD
    flash.position = sprite.position

    target = (sprite.center_x, sprite.center_y + 100)
    seq = Sequence()
    seq[0].frame = KeyFrame.from_sprite(sprite, ['position', 'scale', 'alpha'])
    seq[1].frame = KeyFrame(position=target, scale=1.5, alpha=0)
    seq[1].callback = flash.kill

    arcade.get_window().curtains.current_scene.flash_sprites.append(flash)
    return flash, seq


def attack(sprite):
    x = sprite.center_x
    opx = sprite.opponent.center_x

    seq = Sequence()

    seq[0].frame = KeyFrame.from_sprite(sprite, ['center_x', 'height'])
    seq[.2].frame = KeyFrame(center_x=x - 20, height=sprite.height)
    seq[.3].frame = KeyFrame(center_x=opx, height=sprite.height - 30)
    seq[.31].frame = KeyFrame(center_x=opx, height=sprite.height)
    seq[.8].frame = seq[0].frame

    seq[.3].callback = sprite.opponent.take_damage
    seq[.31].callback = delay_set_attribute(sprite, 'special',
                                            sprite.special + 20)
    seq[.8].callback = delay_set_attribute(sprite, 'stamina', 0)

    return sprite, seq


def takedmg(sprite):
    c = arcade.color
    sprite.color = c.RED

    seq = Sequence()

    seq[0].frame = KeyFrame.from_sprite(sprite, ['center_x'])
    seq[.1].frame = KeyFrame(center_x=sprite.center_x + 30)
    seq[.2].frame = KeyFrame(center_x=sprite.center_x - 10)
    seq[.3].frame = seq[0].frame

    seq[.1].callback = delay_set_attribute(sprite, 'color', sprite.skin_color)
    return sprite, seq


def powerup_generator(sprite):
    rate = sprite.flash_rate
    time = sprite.power_up_time
    flashes = []
    for i in range(int(time / rate)):
        flashes.append(powerup(sprite))

    def make_fn(flash):
        def fn():
            play_animation(flash)

        return fn

    amount = len(flashes)
    for i, (_, seq) in enumerate(flashes):
        if i + 1 < amount:
            seq[rate].callback = make_fn(flashes[i + 1])
        else:
            seq[.99].callback = sprite.attack
    return flashes[0]


class CharacterPortrait(Widget):
    def setup_widget(self, inverse=False):
        c = arcade.color
        bars = [('health', c.RED), ('stamina', c.BLUE), ('special', c.GREEN)]
        self.bars = {}

        for idx, (bar_name, color) in enumerate(reversed(bars)):
            container = Bar(arcade.color.WHITE, height=23, width=103)
            bar = Bar(color, inverse=inverse)
            bar.topleft = (50, idx * 25)
            container.position = bar.position
            self.sprites.extend([container, bar])
            self.bars[bar_name] = bar

        topleft = 0
        if inverse:
            topleft = 121

        padding = CircleSprite(size=80, color=c.SADDLE_BROWN)
        padding.topleft = (topleft, 55)

        portrait = CircleSprite(size=70, color=c.PERU)
        portrait.topleft = (topleft + 5, 50)

        backdrop = SquareSprite()
        backdrop.color = c.SADDLE_BROWN
        backdrop.width = 108
        backdrop.height = 85
        backdrop.position = self.bars['stamina'].position
        backdropleft = CircleSprite(size=85, color=c.SADDLE_BROWN)
        backdropleft.position = (backdrop.left, backdrop.center_y)
        backdropright = CircleSprite(size=85, color=c.SADDLE_BROWN)
        backdropright.position = (backdrop.right, backdrop.center_y)

        for sprite in [backdrop, backdropright, backdropleft]:
            self.sprites.insert(0, sprite)

        self.sprites.extend([padding, portrait])


class Menu(Widget):
    def setup_widget(self, events, hero):
        backdrop = SquareSprite()
        backdrop.color = arcade.color.SADDLE_BROWN
        backdrop.width = 245
        backdrop.height = 95
        backdrop.bottomleft = (0, 0)

        attack = Button(text='Attack', events=events)
        haste = Button(text='Haste', events=events)
        ultimate = Button(text='Ultimate', events=events)
        self.sprites.extend([backdrop, attack, ultimate, haste])

        events.click(attack, hero.attack)
        events.click(ultimate, hero.power_up)
        events.click(haste, hero.haste)

        attack.bottomleft = (5, 50)
        haste.bottomleft = (125, 50)
        ultimate.bottomleft = (5, 5)

        self._sequences = {}
        for sprite in self.sprites:
            sprite.alpha = 0

    @property
    def sequences(self):
        if not self._sequences:
            for sprite in self.sprites:
                seq = Sequence()
                seq[0].frame = KeyFrame.from_sprite(
                    sprite, ['width', 'height', 'position'])
                seq[0].frame.alpha = 255
                seq[.2].frame = KeyFrame(
                    width=0, height=0, position=self.anchor.position, alpha=0)
                self._sequences[sprite] = seq
        return self._sequences

    def show(self):
        for sprite, seq in self.sequences.items():
            seq.is_reversed = True
            sprite.animate(seq)

    def hide(self):
        for sprite, seq in self.sequences.items():
            seq.is_reversed = False
            sprite.animate(seq)


class Ambience(Widget):
    def setup_widget(self, events):
        arcade.set_background_color(arcade.color.SKY_BLUE)

        grass = SquareSprite()
        grass.color = arcade.color.FOREST_GREEN
        grass.width = 500
        grass.height = 200
        grass.bottomleft = (0, 0)
        self.sprites.append(grass)

        class GrassBlade(SquareSprite):
            def __init__(self):
                self.wind = random.random() / 2
                super().__init__()

            def blow_in_the_wind(self, delta):
                self.wind += delta
                if self.wind > 3:
                    self.wind = 0
                elif self.wind > 2.6:
                    self.angle -= delta * 100
                    self.center_x += delta * 10
                elif self.wind > 2.2:
                    self.angle += delta * 100
                    self.center_x -= delta * 10
                elif self.wind > 1.8:
                    self.angle -= delta * 100
                    self.center_x += delta * 10
                elif self.wind > 1.4:
                    self.angle += delta * 100
                    self.center_x -= delta * 10

        for i in range(1100):
            grassblade = GrassBlade()
            grassblade.color = (14, 119, 14)
            grassblade.width = 2
            grassblade.height = 10
            grassblade.bottomleft = (random.randint(0, 500),
                                     random.randint(0, 200))
            self.sprites.append(grassblade)
            events.frame(grassblade.blow_in_the_wind)

        class Cloud(Widget):
            def setup_widget(self):
                base = CircleSprite()
                base.width = 256
                base.height = 64
                base.position = (0, 0)

                top1 = CircleSprite()
                top1.position = (88, 30)

                top2 = CircleSprite()
                top2.scale = 1.3
                top2.position = (20, 50)

                top3 = CircleSprite()
                top3.scale = 0.9
                top3.position = (-50, 20)

                self.sprites.extend([top1, top2, top3, base])
                self.speed = random.randint(50, 100)

            def move(self, delta):
                self.center_x -= delta * self.speed
                if self.center_x < -150:
                    self.center_x = 700

        cloud1 = Cloud()
        cloud1.position = (random.randint(128, 372), 300)
        self.sprites.extend(cloud1.sprites)

        cloud2 = Cloud()
        cloud2.position = (random.randint(128, 372), 400)
        self.sprites.extend(cloud2.sprites)

        cloud3 = Cloud()
        cloud3.position = (random.randint(128, 372), 500)
        self.sprites.extend(cloud3.sprites)

        events.frame(cloud1.move)
        events.frame(cloud2.move)
        events.frame(cloud3.move)


class Controller:
    def __init__(self, char, input_device, scene, inverse):
        self.char = char
        self.input_device = input_device

        portrait = CharacterPortrait(inverse=inverse)
        portrait.position = self.char.position
        portrait.center_y += 240
        portrait.register(scene.sprites)

        self.char.after_change('health', portrait.bars['health'].set_value)
        self.char.after_change('stamina', portrait.bars['stamina'].set_value)
        self.char.after_change('special', portrait.bars['special'].set_value)
        get_state().add_handler(self.char.update)
        char.special = 0

        stamina = TriggerAttr('stamina')
        self.char.trigger(stamina >= 100, self.get_input)
        self.char.trigger(stamina == 0, self.cleanup)

        health = TriggerAttr('health')
        self.char.trigger(health == 0, char.die)

    def get_input(self):
        get_state().set_waiting()
        self.input_device.get_input()

    def cleanup(self):
        get_state().set_idle()
        self.input_device.cleanup()


class Character(SquareSprite, ObservableSprite):
    def __init__(self, name, rate=.02, *args, **kwargs):
        ObservableSprite.__init__(self, *args, **kwargs)
        SquareSprite.__init__(self, *args, **kwargs)
        self.skin_color = arcade.color.PERU
        self.color = self.skin_color
        self.name = name

        self.health = 100
        self.stamina = 0
        self.special = 0
        self.flash_rate = 0.15
        self.power_up_time = 1
        self.opponent = None

        self.recovery_rate = rate
        self.current_recovery_time = 0

        self.damage = 12.5

        self.state = get_state()

    def update(self, delta):
        if self.stamina < 100:
            self.current_recovery_time += delta
            if self.current_recovery_time > self.recovery_rate:
                stam = int(self.current_recovery_time / self.recovery_rate)
                self.current_recovery_time = self.current_recovery_time % self.recovery_rate
                self.stamina += stam

    def power_up(self, *args, **kwargs):
        if self.special < 100:
            return
        play_animation(powerup_generator(self))
        self.special = 0

    def die(self):
        self.kill()
        if self.name == 'hero':
            arcade.get_window().curtains.set_scene('defeat')
        else:
            arcade.get_window().curtains.set_scene('victory')

    def haste(self, *args, **kwargs):
        self.recovery_rate /= 2
        self.stamina = 0

    def attack(self, *args, **kwargs):
        if self.stamina < 100:
            return

        play_animation(attack(self))

    def take_damage(self, *args, **kwargs):
        play_animation(takedmg(self))
        self.health -= self.opponent.damage


class Bar(SquareSprite):
    def __init__(self,
                 color,
                 height=20,
                 width=100,
                 inverse=False,
                 *args,
                 **kwargs):
        SquareSprite.__init__(self, *args, **kwargs)
        self.width = width
        self.height = height
        self.color = color
        self.orientation = 'right' if inverse else 'left'

    def set_perc(self, perc):
        if perc > 1:
            perc = 1
        if perc < 0:
            perc = 0

        orig = getattr(self, self.orientation)
        self.width = 100 * perc
        setattr(self, self.orientation, orig)

    def set_value(self, sprite, attr, old, new):
        if new == 0:
            self.set_perc(0)
        else:
            self.set_perc(new / 100)


class PlayerInput:
    def __init__(self, char, menu):
        self.char = char
        self.menu = menu

    def get_input(self):
        self.menu.show()

    def cleanup(self):
        self.menu.hide()


class AIInput:
    def __init__(self, char):
        self.char = char

    def get_input(self):
        if self.char.special == 100:
            self.char.power_up()
        else:
            self.char.attack()

    def cleanup(self):
        pass


class SmallScene(BaseScene):
    def setup(self):
        self.flash_sprites = arcade.SpriteList()
        self.sprites = arcade.SpriteList()

        ambience = Ambience(events=self.events)
        ambience.bottomleft = (0, 0)
        ambience.register(self.sprites)

        self.hero = Character(name='hero', rate=.019)
        self.hero.position = (100, 200)
        self.villain = Character(name='villain')
        self.villain.position = (400, 200)
        self.hero.opponent = self.villain
        self.villain.opponent = self.hero

        self.sprites.append(self.hero)
        self.sprites.append(self.villain)

        menu = Menu(events=self.events, hero=self.hero)
        menu.register(self.sprites)
        menu.bottomleft = (5, 5)

        Controller(self.hero, PlayerInput(self.hero, menu), self, False)
        Controller(self.villain, AIInput(self.villain), self, True)


class ResultScene(BaseScene):
    def setup(self, message, color):
        self.message = message
        self.color = color
        self.events.after_draw(self.display_message)
        self.alpha = 0

    def enter_scene(self, prev_scene):
        arcade.set_background_color(arcade.color.BLACK)

    def display_message(self):
        self.alpha += 5
        if self.alpha > 255:
            self.alpha = 255
        arcade.draw_text(
            self.message,
            250,
            250,
            self.color + (self.alpha, ),
            35,
            align="center",
            anchor_x="center",
            anchor_y="center")


class Window(arcade.Window):
    def __init__(self):
        super().__init__(500, 500, 'move')
        self.curtains = Curtains(self)
        self.dispatcher = get_state()
        self.curtains.add_scene('scene1', SmallScene())
        self.curtains.add_scene(
            'victory',
            ResultScene(message='You Win :)', color=arcade.color.GREEN))
        self.curtains.add_scene(
            'defeat', ResultScene(
                message='You Lose :(', color=arcade.color.RED))

    def setup(self):
        self.curtains.set_scene('scene1')

    def update(self, delta):
        self.dispatcher.update(delta)


def run(window):
    window.setup()
    try:
        arcade.run()
    except KeyboardInterrupt:
        pass


run(Window())



import arcade
import example_helpers as eh
from arcade_curtains import ObservableSprite
from arcade_curtains.helpers import TriggerAttr


class CircleSprite(eh.CircleSprite, ObservableSprite):
    pass


class Window(arcade.Window):
    def __init__(self):
        super().__init__(500, 500, 'move')
        self.mod = 1
        self.sprites = arcade.SpriteList()
        self.sprite1 = CircleSprite()
        self.sprite1.position = (100, 100)
        self.sprite2 = CircleSprite()
        self.sprite2.position = (100, 400)
        self.sprites.append(self.sprite1)
        self.sprites.append(self.sprite2)

        def keep_pace(sprite, attribute, old, new):
            setattr(self.sprite2, attribute, new)

        def set_mod(mod):
            self.mod = mod

        center_x = TriggerAttr('center_x')
        self.sprite1.trigger(center_x > 400, lambda: set_mod(-1))
        self.sprite1.trigger(center_x < 100, lambda: set_mod(1))
        self.sprite1.after_change('center_x', keep_pace)

    def on_update(self, delta):
        arcade.start_render()
        self.sprite1.center_x += (delta * 100) * self.mod
        self.sprites.draw()


window = Window()
arcade.run()



import arcade

from .event import EventHandler
from .animation import AnimationManager


class BaseScene:
    def __init__(self, *args, **kwargs):
        self.window = None
        self.curtains = None
        self.animations = AnimationManager()
        self.events = EventHandler()
        self.events.frame(self.animations._blip)
        self._sprite_lists = []
        self.setup(*args, **kwargs)
        self._setup_spritelists()

    def _bind(self, window, curtains):
        self.window = window
        self.curtains = curtains

    def setup(self, *args, **kwargs):
        raise NotImplementedError()

    def leave_scene(self, next_scene):
        pass

    def enter_scene(self, previous_scene):
        pass

    def _setup_spritelists(self):
        for attribute in dir(self):
            attr = getattr(self, attribute)
            if isinstance(attr, arcade.SpriteList):
                self._sprite_lists.append(attr)

    def draw(self):
        for slist in self._sprite_lists:
            slist.draw()
